/*
 * Copyright 2013 Red Hat Inc. and/or its affiliates and other contributors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.switchyard.quickstarts.remoteinvoker;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;

import javax.xml.namespace.QName;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.conn.ClientConnectionManager;
import org.apache.http.conn.scheme.PlainSocketFactory;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.entity.ByteArrayEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.PoolingClientConnectionManager;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.CoreConnectionPNames;
import org.apache.http.params.HttpParams;
import org.apache.http.protocol.BasicHttpContext;
import org.apache.http.protocol.HttpContext;
import org.apache.http.util.EntityUtils;
import org.switchyard.quickstarts.bpm.service.data.Order;
import org.switchyard.remote.RemoteMessage;
import org.switchyard.serial.FormatType;
import org.switchyard.serial.Serializer;
import org.switchyard.serial.SerializerFactory;

/**
 * Test client which uses RemoteInvoker to invoke a service with an SCA binding.
 */
public final class RemoteClient {

	private static final QName SERVICE = new QName(
			"urn:switchyard-quickstart:bpm-service:1.0", "Dealer");
	private static final QName SERVICE_TEST = new QName(
			"urn:switchyard-quickstart:bpm-service:1.0", "ProcessOrder");
	private static final String URL = "http://localhost:8080/switchyard-remote";
	public static final String SERVICE_HEADER = "switchyard-service";

	/**
	 * Private no-args constructor.
	 */
	private RemoteClient() {
	}

	/**
	 * Only execution point for this application.
	 * 
	 * @param ignored
	 *            not used.
	 * @throws Exception
	 *             if something goes wrong.
	 */

	private static HttpClient client;
	private static IdleConnectionMonitorThread idleEvictor;
	private static AtomicLong cnt = new AtomicLong(0L);

	public static void main(final String[] ignored) throws Exception {

		SchemeRegistry schemeRegistry = new SchemeRegistry();
		schemeRegistry.register(new Scheme("http", 80, PlainSocketFactory
				.getSocketFactory()));
		schemeRegistry.register(new Scheme("https", 443, SSLSocketFactory
				.getSocketFactory()));

		HttpParams params = new BasicHttpParams();
		params.setParameter(CoreConnectionPNames.SO_TIMEOUT, 0);
		params.setParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, 30000);
		params.setParameter(CoreConnectionPNames.STALE_CONNECTION_CHECK, true);

		PoolingClientConnectionManager cm = new PoolingClientConnectionManager(
				schemeRegistry);
		cm.setMaxTotal(800);
		cm.setDefaultMaxPerRoute(400);

		idleEvictor = new IdleConnectionMonitorThread(cm, 30000, 5000);
		idleEvictor.start();

		client = new DefaultHttpClient(cm);

		ArrayList<Thread> threads = new ArrayList<Thread>();

		int threadCount = 50;
		final int count = 2000000;

		for (int j = 0; j < threadCount; ++j) {
			threads.add(new Thread(new Runnable() {

				@Override
				public void run() {
					Serializer _serializer = SerializerFactory.create(
							FormatType.JSON, null, true);
					HttpContext ctx = new BasicHttpContext();
					ByteArrayOutputStream os = new ByteArrayOutputStream();
					ByteArrayOutputStream osResp = new ByteArrayOutputStream();
					for (int j = 0; j < count; ++j) {
						RemoteMessage reply = null;
						try {
							HttpPost httpRequest = new HttpPost(URL);
							// Create request payload
							// Offer offer = createOffer(true);
							Order order = new Order();
							order.setItemId("cowbell");
							order.setOrderId("SHIPIT");
							order.setQuantity(50);
							// Create the request message
							RemoteMessage message = new RemoteMessage();
							message.setService(SERVICE_TEST)
									.setOperation("submitOrder")
									.setContent(order);
							httpRequest.setHeader(SERVICE_HEADER, message
									.getService().toString());
							os.reset();
							_serializer.serialize(message, RemoteMessage.class,
									os);
							os.flush();
							httpRequest.setEntity(new ByteArrayEntity(os
									.toByteArray()));
							HttpResponse resp = client
									.execute(httpRequest, ctx);
							if (resp.getStatusLine().getStatusCode() == 200) {
								HttpEntity entity = resp.getEntity();
								InputStream is = entity.getContent();
								byte[] buffer = new byte[4096];
								int len = -1;
								osResp.reset();
								while ((len = is.read(buffer)) != -1) {
									osResp.write(buffer, 0, len);
								}
								osResp.flush();
								ByteArrayInputStream bais = new ByteArrayInputStream(
										osResp.toByteArray());
								reply = _serializer.deserialize(bais,
										RemoteMessage.class);
								EntityUtils.consume(entity);
							}
						} catch (Throwable t) {
							t.printStackTrace();
							try {
								System.out.println(new String(osResp
										.toByteArray(), "UTF-8"));
							} catch (UnsupportedEncodingException e) {
								e.printStackTrace();
							}
						}
						if (reply == null) {
							System.out.println("NULL reply!");
						}
						long sss;
						if ((sss = cnt.incrementAndGet()) % 1000 == 0) {
							System.out.println(sss);
							try {
								System.out.println("Last reply: "
										+ new String(osResp.toByteArray(),
												"UTF-8"));
							} catch (UnsupportedEncodingException e) {

							}
						}
					}
				}
			}));
		}

		long startTime = System.currentTimeMillis();

		for (Thread thread : threads) {
			thread.start();
		}

		for (Thread thread : threads) {
			try {
				thread.join();
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		long endTime = System.currentTimeMillis();

		float tps = (((float) (count * threadCount)) / (endTime - startTime)) * 1000;

		System.out.println("Tps = " + tps);

		// RemoteMessage reply = invoker.invoke(message);
		// if (reply.isFault()) {
		// System.err.println("Oops ... something bad happened.  "
		// + reply.getContent());
		// } else {
		// Deal deal = (Deal) reply.getContent();
		// out.println("==================================");
		// out.println("Was the offer accepted? " + deal.isAccepted());
		// out.println("==================================");
		// }
		idleEvictor.shutdown();
	}

	public static Offer createOffer(boolean acceptable) {
		Application app = new Application();
		app.setName("John Smith");
		app.setCreditScore(acceptable ? 700 : 300);
		Car car = new Car();
		car.setPrice(18000);
		car.setVehicleId("Honda");
		Offer offer = new Offer();
		offer.setApplication(app);
		offer.setCar(car);
		offer.setAmount(17000);

		return offer;
	}

	public static class IdleConnectionMonitorThread extends Thread {

		private final ClientConnectionManager connMgr;
		private final int idleTimeout;
		private final int idleMonitorInterval;

		private volatile boolean active;

		public IdleConnectionMonitorThread(ClientConnectionManager connMgr,
				int idleTimeout, int idleMonitorInterval) {
			super();
			this.connMgr = connMgr;
			this.idleTimeout = idleTimeout;
			this.idleMonitorInterval = idleMonitorInterval;
			active = true;
		}

		@Override
		public void run() {
			while (active) {
				try {
					synchronized (this) {
						wait(idleMonitorInterval);
						connMgr.closeExpiredConnections();
						connMgr.closeIdleConnections(idleTimeout,
								TimeUnit.MILLISECONDS);
					}
				} catch (InterruptedException ex) {
					// ignore
				}
			}
		}

		public void shutdown() {
			active = false;
			synchronized (this) {
				notifyAll();
			}
		}
	}
}
